/* Copyright 2002 Andi Kleen */

#include <linux/linkage.h>

#include <asm/cpufeature.h>
#include <asm/dwarf2.h>

/*
 * memcpy - Copy a memory block.
 *
 * Input:
 *  rdi destination
 *  rsi source
 *  rdx count
 *
 * Output:
 * rax original destination
 */

ENTRY(__new_memcpy)
	CFI_STARTPROC
	movq %rdi, %rax

	/*
	 * Use 32bit CMP here to avoid long NOP padding.
	 */
	cmp  $0x20, %edx
	jbe .Lhandle_tail

	/*
	 * the code for unaligned copy is good for large-size copy(>100),
	 * so if the size is small, we needn't check dst and src is aligned
	 * or not.
	 */
	cmp $100, %edx
	jb .Lboth_aligned

	/*
	 * unaligned access always leads to bad performance, so in order to
	 * avoid unaligned access, we align the address(both src and dest)
	 * first, and then copy from a aligned src to an aligned dst by using
	 * shifts.
	 * But we found if src is aligned, although dest is unaligned, the
	 * performance of generic memory copy (That is reading data aligned
	 * from the source and writing data unaligned to the dest) is better
	 * than the one that uses shifts to avoid unaligned access.
	 * So if src is aligned, we needn't check dest is aligned or not, just
	 * goto .Lboth_aligned
	 */
	test $7, %esi		/* src align check */
	jz .Lboth_aligned

	/* if dest and src both are unaligned, goto unaligned copy */
	test $7, %edi
	jnz .Ldst_unaligned

	jmp .Lsrc_unaligned_dst_aligned

.Lboth_aligned:
	/*
	 * We check whether memory false dependece could occur,
	 * then jump to corresponding copy mode.
	 */
	cmp  %dil, %sil
	jl .Lcopy_backward
	subl $0x20, %edx
.Lcopy_forward_loop:
	subq $0x20,	%rdx

	/*
	 * Move in blocks of 4x8 bytes:
	 */
	movq 0*8(%rsi),	%r8
	movq 1*8(%rsi),	%r9
	movq 2*8(%rsi),	%r10
	movq 3*8(%rsi),	%r11
	leaq 4*8(%rsi),	%rsi

	movq %r8,	0*8(%rdi)
	movq %r9,	1*8(%rdi)
	movq %r10,	2*8(%rdi)
	movq %r11,	3*8(%rdi)
	leaq 4*8(%rdi),	%rdi
	jae  .Lcopy_forward_loop
	addq $0x20,	%rdx
	jmp  .Lhandle_tail

.Lcopy_backward:
	/*
	 * Calculate copy position to tail.
	 */
	addq %rdx,	%rsi
	addq %rdx,	%rdi
	subq $0x20,	%rdx
	/*
	 * At most 3 ALU operations in one cycle,
	 * so append NOPS in the same 16bytes trunk.
	 */
	.p2align 4
.Lcopy_backward_loop:
	subq $0x20,	%rdx
	movq -1*8(%rsi),	%r8
	movq -2*8(%rsi),	%r9
	movq -3*8(%rsi),	%r10
	movq -4*8(%rsi),	%r11
	leaq -4*8(%rsi),	%rsi
	movq %r8,		-1*8(%rdi)
	movq %r9,		-2*8(%rdi)
	movq %r10,		-3*8(%rdi)
	movq %r11,		-4*8(%rdi)
	leaq -4*8(%rdi),	%rdi
	jae  .Lcopy_backward_loop

	/*
	 * Calculate copy position to head.
	 */
	addq $0x20,	%rdx
	subq %rdx,	%rsi
	subq %rdx,	%rdi
.Lhandle_tail:
	cmpq $16,	%rdx
	jb   .Lless_16bytes

	/*
	 * Move data from 16 bytes to 31 bytes.
	 */
	movq 0*8(%rsi), %r8
	movq 1*8(%rsi),	%r9
	movq -2*8(%rsi, %rdx),	%r10
	movq -1*8(%rsi, %rdx),	%r11
	movq %r8,	0*8(%rdi)
	movq %r9,	1*8(%rdi)
	movq %r10,	-2*8(%rdi, %rdx)
	movq %r11,	-1*8(%rdi, %rdx)
	retq
	.p2align 4
.Lless_16bytes:
	cmpq $8,	%rdx
	jb   .Lless_8bytes
	/*
	 * Move data from 8 bytes to 15 bytes.
	 */
	movq 0*8(%rsi),	%r8
	movq -1*8(%rsi, %rdx),	%r9
	movq %r8,	0*8(%rdi)
	movq %r9,	-1*8(%rdi, %rdx)
	retq
	.p2align 4
.Lless_8bytes:
	cmpq $4,	%rdx
	jb   .Lless_3bytes

	/*
	 * Move data from 4 bytes to 7 bytes.
	 */
	movl (%rsi), %ecx
	movl -4(%rsi, %rdx), %r8d
	movl %ecx, (%rdi)
	movl %r8d, -4(%rdi, %rdx)
	retq
	.p2align 4
.Lless_3bytes:
	cmpl $0, %edx
	je .Lend
	/*
	 * Move data from 1 bytes to 3 bytes.
	 */
.Lloop_1:
	movb (%rsi), %r8b
	movb %r8b, (%rdi)
	incq %rdi
	incq %rsi
	decl %edx
	jnz .Lloop_1

.Lend:
	retq

	.p2align 4
.Ldst_unaligned:
	movq %rdi, %rcx
	andq $7, %rcx		/* Align the destination */
	negq %rcx
	andq $7, %rcx
	subq %rcx, %rdx

	/* tune dst address */
	movq (%rsi), %r8
	movq %r8, (%rdi)
	addq %rcx, %rdi
	addq %rcx, %rsi

	test $7, %esi		/* src align check */
	jz .Lboth_aligned

	.p2align 4
.Lsrc_unaligned_dst_aligned:
	push %rbx
	push %r12
	push %r13
	push %r14
	push %r15
	/*
	 * Calculate how to shift a word read at the memory operation
	 * aligned srcp to make it aligned for copy.
	 */
	movq %rsi, %r14
	andq $7, %r14
	shlq $3, %r14
	
	movq $64, %r15
	subq %r14, %r15

	andq $-8, %rsi		/* src aligned */
	movq 0*8(%rsi), %r8

	movq %rdx, %rbx
	shrq $5, %rbx
	jz .Lsrc_unaligned_less32

	/*
	 * %r8 : store src[0]
	 * %r9 : store src[1]
	 * %r10: store src[2]
	 * %r11: store src[3]
	 * %r12: store src[4]
	 * %r13: store the tmp data
	 */ 
	.p2align 4
.Lsrc_unaligned_loop32:
	movq 1*8(%rsi), %r9
	movq 2*8(%rsi), %r10
	movq 3*8(%rsi), %r11
	movq 4*8(%rsi), %r12

	movq %r9, %r13
	movb %r14b, %cl
	shrq %cl, %r8
	shrq %cl, %r13
	movb %r15b, %cl
	shlq  %cl, %r9
	orq %r8, %r9
	movq %r10, %r8
	shlq  %cl, %r10
	orq %r13, %r10

	movq %r11, %r13
	movb %r14b, %cl
	shrq %cl, %r8
	shrq %cl, %r13
	movb %r15b, %cl
	shlq  %cl, %r11
	orq %r8, %r11	
	movq %r12, %r8
	shlq  %cl, %r12
	orq %r13, %r12

	movq %r9, 0*8(%rdi)
	movq %r10, 1*8(%rdi)
	movq %r11, 2*8(%rdi)
	movq %r12, 3*8(%rdi)
	
	leaq 4*8(%rdi), %rdi
	leaq 4*8(%rsi), %rsi
	decq %rbx
	jnz .Lsrc_unaligned_loop32

	.p2align 4
.Lsrc_unaligned_less32:
	shrq $3, %r14
	addq %r14, %rsi
	pop %r15
	pop %r14
	pop %r13
	pop %r12
	pop %rbx
	andq $31, %rdx
	jnz .Lhandle_tail
	retq
	
	CFI_ENDPROC
ENDPROC(__new_memcpy)

